package impl.evaluation;

import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpHandler;
import com.sun.net.httpserver.HttpServer;
import impl.indexing.CorpusIndexer;
import impl.indexing.parsing.CorpusParser;
import impl.language.Language;
import impl.language.Translator;
import impl.model.Query;
import impl.model.SearchReport;

import java.io.*;
import java.net.BindException;
import java.net.InetSocketAddress;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

public class WebInterface {

	SearchService searchService;
	CorpusIndexer corpusIndexer;

	public static void main(String[] args) {
		new WebInterface();
	}

	public WebInterface(SearchService searchService, CorpusIndexer corpusIndexer) {
		this.searchService = searchService;
		this.corpusIndexer = corpusIndexer;
		beginServer();
	}

	public WebInterface() {
		searchService = null;
		corpusIndexer = null;
		beginServer();
	}

	private void beginServer() {
		int port = 8000;
		for(;;) {
			try {
				HttpServer hs = HttpServer.create(new InetSocketAddress(port), 0);
				hs.createContext("/", new TestHandler());
				hs.setExecutor(null);
				System.out.println("Web server started at http://127.0.0.1:" + port + "/");
				hs.start();
				break;
			} catch (BindException e) {
				System.err.println("Port " + port + " already in use.");
				port++;
			} catch (IOException e) {
				e.printStackTrace();
				break;
			}
		}
	}

	private class TestHandler implements HttpHandler {

		public void handle(HttpExchange he) throws IOException {
			try {
				System.out.println(he.getRequestMethod() + " " + he.getRequestURI().toString());
				String path = he.getRequestURI().getPath();
				if (path == null || !path.startsWith("/")) {
					sendString(he, 400, "Bad request");
				} else if (path.startsWith("/ajax")) {
					handleAjax(he);
				} else if (path.startsWith("/jquery/")) {
					URL lib = getClass().getResource("/META-INF/resources/webjars/jquery/2.1.4" + path.substring(7));
					if (lib == null) {
						sendString(he, 403, "Could not locate jquery library");
					} else {
						sendString(he, 200, openURL(lib));
					}
				} else if (path.equals("/")) {
					sendFile(he, new File("html/index.html"));
				} else if (path.startsWith("/corpus/")) {
					retrieveCorpus(he);
				} else {
					sendFile(he, new File("html" + path));
				}
			} catch (Exception e) {
				e.printStackTrace();
				sendString(he, 500, e.getClass().getName());
			}
		}

		private int handleAjax(HttpExchange he) throws IOException {
			if (searchService == null) {
				return sendString(he, 501, "The web server has been started directly without reference to any database.");
			}
			Map<String, String> queryParams = splitQuery(he.getRequestURI().getQuery());
			if (!queryParams.containsKey("query")) {
				return sendString(he, 400, "Query is missing");
			}
			Language sourceLanguage = getLanguageByParam(queryParams, "lang_src");
			if (sourceLanguage == null) {
				return sendString(he, 400, "Source language is missing");
			}
			Language targetLanguage = getLanguageByParam(queryParams, "lang_trg");
			Query query = Query.byString(queryParams.get("query"), sourceLanguage);
			SearchReport sr = searchService.perform(query, targetLanguage, true, Translator.Pruning.BALANCE);
			//System.out.println(sr.toJson());
			return sendString(he, 200, sr.toJson());
		}

		private Language getLanguageByParam(Map<String, String> queryParams, String paramName) {
			if (!queryParams.containsKey(paramName)) {
				return null;
			}
			try {
				return Language.valueOf(queryParams.get(paramName));
			} catch(IllegalArgumentException e) {
				return null;
			}
		}

		private int retrieveCorpus(HttpExchange he) throws Exception {
			if (corpusIndexer == null) {
				return sendString(he, 501, "The web server has been started directly without reference to any corpora.");
			}
			String path = he.getRequestURI().getPath();
			String title = path.substring(8);
			int pos = title.indexOf("/");
			if (pos == -1) {
				return corpusRequestError(he);
			}
			String source = title.substring(0, pos);
			title = title.substring(pos + 1);
			pos = title.indexOf("/");
			if (pos == -1) {
				return corpusRequestError(he);
			}
			String collection = title.substring(0, pos);
			title = title.substring(pos + 1);
			pos = title.indexOf("/");
			if (pos == -1) {
				return corpusRequestError(he);
			}
			String langName = title.substring(0, pos).toUpperCase();
			title = title.substring(pos + 1);
			if ("".equals(title)) {
				return corpusRequestError(he);
			}
			Language language = Language.valueOf(langName);
			CorpusParser.RawCorpusAccess rca = corpusIndexer.getRawCorpusAccess(source, collection, language, title);
			if (rca == null) {
				return sendString(he, 404, "Not found");
			}
			else {
				if (rca.mime != null) {
					he.getResponseHeaders().add("Content-type", rca.mime);
				}
				return sendStream(he, 200, rca.stream, rca.length);
			}
		}

		private int corpusRequestError(HttpExchange he) throws IOException {
			return sendString(he, 400, "Request has to be formatted as: /source/collection/language/title");
		}

		private int sendFile(HttpExchange he, File file) throws IOException {
			if (!file.canRead()) {
				if (!file.exists()) {
					return sendString(he, 404, "Not found");
				}
				return sendString(he, 403, "Access denied");
			}
			return sendStream(he, 200, new FileInputStream(file), file.length());
		}

		private int sendString(HttpExchange he, int code, String string) throws IOException {
			return sendStream(he, code, new ByteArrayInputStream(string.getBytes()), string.length());
		}

		private int sendStream(HttpExchange he, int code, InputStream is, long length) throws IOException {
			he.sendResponseHeaders(code, length);
			OutputStream os = he.getResponseBody();
			long left = length;
			byte[] buffer = new byte[(int) Math.min(left, 4096)];
			while(left > 0 && is.read(buffer) > -1) {
				os.write(buffer);
				left -= 4096;
				buffer = new byte[(int) Math.max(0, Math.min(left, 4096))];
			}
			is.close();
			os.close();
			System.out.println("HTTP " + code);
			return code == 200 ? 0 : 1;
		}

		private Map<String, String> splitQuery(String input) {
			Map<String, String> query = new HashMap<String, String>();
			String[] split;
			for (String part : input.split("&")) {
				split = part.split("=");
				if (split.length == 2) {
					query.put(split[0], split[1]);
				}
			}
			return query;
		}

		private String openURL(URL u) throws IOException {
			InputStream is = u.openStream();
			StringBuilder sb = new StringBuilder();
			InputStreamReader isr = new InputStreamReader(is);
			BufferedReader br = new BufferedReader(isr);
			String line;
			while ((line = br.readLine()) != null) {
				sb.append(line);
			}
			br.close();
			isr.close();
			is.close();
			return sb.toString();
		}

	}

}
