package impl.indexing.extracting;

import impl.language.Language;
import impl.model.Environment;
import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveInputStream;
import org.apache.commons.compress.archivers.tar.TarArchiveOutputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorInputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorOutputStream;
import org.apache.commons.compress.utils.IOUtils;

import java.io.*;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class OpusOpenSubtitlesPruner {
	private static int maxPar = 500;
	private static int maxExt = 2000000000;

	private String source;
	private String target;
	private String extDir;
	private Edition edition;

	private Map<String, Set<Language>> documents;
	private Set<Language> languages;
	private Set<String> parallels;
	private Map<Language, Set<String>> externals;

	private OpusOpenSubtitlesPruner(String source, String target, String extDir, Edition edition) {
		this.source = source;
		this.target = target;
		this.extDir = extDir;
		this.edition = edition;
		documents = new HashMap<String, Set<Language>>();
		languages = new HashSet<Language>();
		parallels = new HashSet<String>();
		externals = new HashMap<Language, Set<String>>();
	}

	public static void main(String[] args) {
		OpusOpenSubtitlesPruner oosp = /*new OpusOpenSubtitlesPruner(
				Environment.getInstance().getProperty("opusCorporaDirectory") + "/OpenSubtitles/",
				Environment.getInstance().getProperty("opusParallelCorporaDirectory") + "/OpenSubtitles/",
				Environment.getInstance().getProperty("opusExternalCorporaDirectory") + "/OpenSubtitles/",
				Edition.OS
		);
		oosp.prune();
		oosp = */new OpusOpenSubtitlesPruner(
				Environment.getInstance().getProperty("opusCorporaDirectory") + "/OpenSubtitles2016/",
				Environment.getInstance().getProperty("opusParallelCorporaDirectory") + "/OpenSubtitles2016/",
				Environment.getInstance().getProperty("opusExternalCorporaDirectory") + "/OpenSubtitles2016/",
				Edition.OS2016
		);
		oosp.prune();
	}

	private void prune() {
		readCollection(new File(source));
		for (Map.Entry<String, Set<Language>> found : documents.entrySet()) {
			languages.addAll(found.getValue());
			//System.out.println(found.getKey() + found.getValue());
		}
		String file;
		for (Language lang : languages) {
			externals.put(lang, new HashSet<String>());
			file = lang.toString().toLowerCase() + ".tar.gz";
			copyTarball(lang, source + file, target + file, extDir + file);
		}
		int count = 0;
		for (Set<Language> found : documents.values()) {
			if (found.size() == languages.size()) {
				count++;
			}
		}
		System.out.println("Total parallel documents: " + count + " (" + parallels.size() + " used)");
		for (Language lang : languages) {
			System.out.println("Leftover documents used as external corpora[" + lang + "]: " + externals.get(lang).size());
		}
	}

	private void readCollection(File collectionDir) {
		System.out.println(collectionDir.getAbsolutePath());
		for (File languageFile : collectionDir.listFiles()) {
			if (!languageFile.isFile() || !languageFile.getName().endsWith(".tar.gz") || languageFile.getName().length() != 9) {
				continue;
			}
			readLanguageFile(languageFile);
		}
	}

	private void readLanguageFile(File languageFile) {
		String shortName = languageFile.getName();
		shortName = shortName.substring(0, shortName.indexOf(".")).toUpperCase();
		Language language = Language.valueOf(shortName);
		try {
			FileInputStream fis = new FileInputStream(languageFile);
			GzipCompressorInputStream gis = new GzipCompressorInputStream(fis);
			readTarball(language, gis);
			gis.close();
			fis.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void readTarball(Language language, GzipCompressorInputStream gis) {
		TarArchiveInputStream tis = new TarArchiveInputStream(gis);
		try {
			TarArchiveEntry tae;
			while((tae = tis.getNextTarEntry()) != null) {
				if (tae.getName().length() >= 100) {
					continue;
				}
				registerDocument(simplifyTitle(tae.getName()), language);
			}
			tis.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void copyTarball(Language language, String srcPath, String trgPath, String extPath) {
		try {
			File out1 = new File(trgPath);
			if (!out1.exists()) {
				(new File(out1.getParent())).mkdirs();
				out1.createNewFile();
			}
			File out2 = new File(extPath);
			if (!out2.exists()) {
				(new File(out2.getParent())).mkdirs();
				out2.createNewFile();
			}
			FileInputStream fis = new FileInputStream(srcPath);
			FileOutputStream fos1 = new FileOutputStream(trgPath);
			FileOutputStream fos2 = new FileOutputStream(extPath);
			GzipCompressorInputStream gis = new GzipCompressorInputStream(fis);
			GzipCompressorOutputStream gos1 = new GzipCompressorOutputStream(fos1);
			GzipCompressorOutputStream gos2 = new GzipCompressorOutputStream(fos2);
			TarArchiveInputStream tis = new TarArchiveInputStream(gis);
			TarArchiveOutputStream tos1 = new TarArchiveOutputStream(gos1);
			TarArchiveOutputStream tos2 = new TarArchiveOutputStream(gos2);
			TarArchiveEntry tae;
			String title;
			while((tae = tis.getNextTarEntry()) != null) {
				if (tae.getName().length() >= 100) {
					continue;
				}
				title = simplifyTitle(tae.getName());
				if (documents.get(title).size() < languages.size() || (maxPar <= parallels.size() && !parallels.contains(title))) {
				//if (!documents.get(title).contains(Language.NO)) {
					if (externals.get(language).size() < maxExt && !externals.get(language).contains(title)) {
						tos2.putArchiveEntry(tae);
						IOUtils.copy(tis, tos2);
						tos2.closeArchiveEntry();
						externals.get(language).add(title);
					}
					continue;
				}
				parallels.add(title);
				tos1.putArchiveEntry(tae);
				IOUtils.copy(tis, tos1);
				tos1.closeArchiveEntry();
			}
			tis.close();
			tos1.close();
			tos2.close();
			gis.close();
			gos1.close();
			gos2.close();
			fis.close();
			fos1.close();
			fos2.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void registerDocument(String title, Language language) {
		if (!documents.containsKey(title)) {
			documents.put(title, new HashSet<Language>());
		}
		documents.get(title).add(language);
	}

	private String simplifyTitle(String title) {
		switch (edition) {
			case OS:
				return title
						.replaceFirst("\\/[a-z]{2}\\/", "/")
						.replaceFirst("\\/[\\d]+_[\\d]+_[\\d]+_(.*?).xml.gz", "/$1");
			case OS2016:
				return title
						.replaceFirst("\\/[a-z]{2}\\/", "/")
						.replaceFirst("\\/[\\d]+.xml.gz", "");
			default:
				return null;
		}
	}

	private enum Edition {
		OS, OS2016
	}
}
