package impl.indexing.parsing;

import impl.language.Language;
import impl.language.Preprocessor;
import impl.language.splitting.Splitter;
import impl.model.*;
import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveInputStream;
import org.apache.commons.compress.archivers.zip.ZipArchiveEntry;
import org.apache.commons.compress.archivers.zip.ZipFile;
import org.apache.commons.compress.compressors.gzip.GzipCompressorInputStream;

import java.io.*;
import java.nio.charset.Charset;
import java.util.*;

public class AvisCorusParser implements CorpusParser {
	private File sourceFolder;
	private CorpusService corpusService;
	private boolean shouldRun;
	private Preprocessor preprocessor;
	private int counter;
	private int maxIndexed;
	private Map<Language, Integer> indexed;
	private Splitter splitter;

	public AvisCorusParser() {
		shouldRun = true;
		preprocessor = new Preprocessor(Language.NO);
		counter = 0;
		maxIndexed = -1;
		splitter = null;
	}

	public AvisCorusParser(Splitter splitter) {
		shouldRun = true;
		preprocessor = new Preprocessor(Language.NO);
		counter = 0;
		this.maxIndexed = -1;
		indexed = new HashMap<Language, Integer>();
		this.splitter = splitter;
	}

	public void setSource(Object source) throws IllegalArgumentException {
		if (source == null || !(source instanceof File) || !((File) source).isDirectory()) {
			throw new IllegalArgumentException("Only directories are supported by the parser.");
		}
		sourceFolder = (File) source;
	}

	public void setCorpusService(CorpusService cs) {
		corpusService = cs;
	}

	public void indexFromSource() throws IllegalStateException {
		if (sourceFolder == null) {
			throw new IllegalStateException("No source is provided.");
		}
		if (corpusService == null) {
			throw new IllegalStateException("CorpusService is not provided.");
		}
		readSource();
	}

	public RawCorpusAccess getRawCorpusAccess(String searchCollection, Language language, String title) throws Exception {
		if (corpusService == null) {
			throw new Exception("CorpusService is not provided.");
		}
		return null;//searchSource(searchCollection, language, title);
	}

	public void terminate() {
		shouldRun = false;
		System.out.println("Terminating Avis parser...");
	}

	public void readSource() {
		List<File> files = Arrays.asList(sourceFolder.listFiles());
		//Collections.reverse(files);
		for (File zipFile : files) {
			if (!shouldRun) {
				System.out.println("Terminated Avis parser");
				return;
			}
			if (!zipFile.isFile() || !zipFile.getName().endsWith(".zip")) {
				continue;
			}
			readZipFile(zipFile.getName(), zipFile);
		}
	}

	private void readZipFile(String collection, File zipFile) {
		try {
			ZipFile zf = new ZipFile(zipFile);
			Enumeration<ZipArchiveEntry> ze = zf.getEntries();
			ZipArchiveEntry zae;
			InputStream zis;
			while (ze.hasMoreElements()) {
				if (!shouldRun) {
					System.out.println("Terminated Avis parser");
					return;
				}
				zae = ze.nextElement();
				//System.out.println(zae.getName());
				if (zae.getName().startsWith("1/") || zae.isDirectory()) {
					continue;
				}
				zis = zf.getInputStream(zae);
				readTarball(collection, zae.getName(), zis);
				//zis.close();
			}
			zf.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void readTarball(String collection, String category, InputStream zis) {
		try {
			GzipCompressorInputStream gis = new GzipCompressorInputStream(zis);
			TarArchiveInputStream tis = new TarArchiveInputStream(gis);
			TarArchiveEntry tae;
			byte[] data;
			while((tae = tis.getNextTarEntry()) != null) {
				if (!shouldRun) {
					System.out.println("Terminated Avis parser");
					return;
				}
				if (tae.isDirectory()) {
					continue;
				}
				if (corpusService.exists(collection, Language.NO, category + "/" + tae.getName())) {
					System.err.println("Skipping: " + tae.getName() + " " + collection);
				}
				else {
					//System.out.println("\t" + tae.getName());
					data = new byte[(int) tae.getSize()];
					DataInputStream dis = new DataInputStream(tis);
					dis.read(data);
					readCorpus(collection, category, tae.getName(), data);
					//Thread.sleep(500);
				}
			}
			gis.close();
			tis.close();
		} catch (IOException e) {
			e.printStackTrace();
			//} catch (InterruptedException e) {
			//	e.printStackTrace();
		}
	}

	private void readCorpus(String collection, String category, String title, byte[] data) {
		@SuppressWarnings("Since15") String text = new String(data, Charset.forName("ISO-8859-1"));
		System.out.println(text.length());
		//System.out.println(text);
		if (text.indexOf("<html") != -1) {
			text = Useful.stripElements(text.toString());
			//System.out.println(text);
		}
		Corpus instance = new Corpus(collection, Language.NO, category + "/" + title, "Avis");
		List<String> tokens = Arrays.asList(text.split("\\s+"));
		if (splitter != null) {
			tokens = splitter.process(tokens);
		}
		List<Term> terms = preprocessor.process(tokens);
		//System.out.println(terms);
		instance.addTerms(terms);
		instance.setId(corpusService.addCorpus(instance));
		System.out.println(instance);
	}
}
