package impl.indexing.parsing;

import impl.evaluation.SearchService;
import impl.language.Language;
import impl.language.Preprocessor;
import impl.language.splitting.RanangSplitter;
import impl.language.splitting.Splitter;
import impl.language.splitting.StatisticalSplitter;
import impl.model.*;
import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveInputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorInputStream;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;

import java.io.*;
import java.util.*;

public class OpusCorpusParser implements CorpusParser {
	private File sourceFolder;
	private CorpusService corpusService;
	private boolean shouldRun;
	private Map<Language, Preprocessor> preprocessors;
	private int counter;
	private int maxIndexed;
	private Map<Language, Integer> indexed;
	private Database external;
	private Map<Language, Splitter> splitter;
	private Splitter ranangSplitter;

	public OpusCorpusParser() {
		shouldRun = true;
		preprocessors = new HashMap<Language, Preprocessor>();
		counter = 0;
		maxIndexed = -1;
		indexed = new HashMap<Language, Integer>();
	}

	public OpusCorpusParser(int maxIndexed) {
		shouldRun = true;
		preprocessors = new HashMap<Language, Preprocessor>();
		counter = 0;
		this.maxIndexed = maxIndexed;
		indexed = new HashMap<Language, Integer>();
	}

	public OpusCorpusParser(Database external, boolean ranang) {
		shouldRun = true;
		preprocessors = new HashMap<Language, Preprocessor>();
		counter = 0;
		this.maxIndexed = -1;
		indexed = new HashMap<Language, Integer>();
		this.external = external;
		if (ranang) {
			LexiconService ls = new LexiconService(external);
			ranangSplitter = new RanangSplitter(ls.getLexicon(Language.NO));
		}
		else {
			ranangSplitter = null;
		}
	}

	public void setSource(Object source) throws IllegalArgumentException {
		if (source == null || !(source instanceof File) || !((File) source).isDirectory()) {
			throw new IllegalArgumentException("Only directories are supported by the parser.");
		}
		sourceFolder = (File) source;
	}

	public void setCorpusService(CorpusService cs) {
		corpusService = cs;
	}

	public void indexFromSource() throws IllegalStateException {
		if (sourceFolder == null) {
			throw new IllegalStateException("No source is provided.");
		}
		if (corpusService == null) {
			throw new IllegalStateException("CorpusService is not provided.");
		}
		readSource();
	}

	public RawCorpusAccess getRawCorpusAccess(String searchCollection, Language language, String title) throws Exception {
		//if (corpusService == null) {
		//	throw new Exception("CorpusService is not provided.");
		//}
		return searchSource(searchCollection, language, title);
	}

	public void terminate() {
		shouldRun = false;
		System.out.println("Terminating Opus parser...");
	}

	public void readSource() {
		for (File collectionDir : sourceFolder.listFiles()) {
			if (!collectionDir.isDirectory()) {
				continue;
			}
			readCollection(collectionDir);
		}
	}

	private void readCollection(File collectionDir) {
		List<File> files = Arrays.asList(collectionDir.listFiles());
		Collections.reverse(files);
		for (File languageFile : files) {
			if (!shouldRun) {
				System.out.println("Terminated Opus parser");
				return;
			}
			if (!languageFile.isFile() || !languageFile.getName().endsWith(".tar.gz")) {
				continue;
			}
			readLanguageFile(collectionDir.getName(), languageFile);
		}
	}

	private void readLanguageFile(String collection, File languageFile) {
		String shortName = languageFile.getName();
		shortName = shortName.substring(0, shortName.indexOf(".")).toUpperCase();
		Language language = Language.valueOf(shortName);
		try {
			indexed.put(language, 0);
			FileInputStream fis = new FileInputStream(languageFile);
			GzipCompressorInputStream gis = new GzipCompressorInputStream(fis);
			readTarball(collection, language, gis);
			gis.close();
			fis.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void readTarball(String collection, Language language, GzipCompressorInputStream gis) {
		TarArchiveInputStream tis = new TarArchiveInputStream(gis);
		try {
			TarArchiveEntry tae;
			int num;
			byte[] data;
			while((tae = tis.getNextTarEntry()) != null) {
				if (!shouldRun) {
					System.out.println("Terminated Opus parser");
					return;
				}
				if (corpusService.exists(collection, language, tae.getName())) {
					System.err.println("Skipping: " + tae.getName() + " " + language + "/" + collection);
				}
				else {
					data = new byte[(int) tae.getSize()];
					DataInputStream dis = new DataInputStream(tis);
					dis.read(data);
					ByteArrayInputStream bis = new ByteArrayInputStream(data);
					GzipCompressorInputStream gis2 = new GzipCompressorInputStream(bis);
					readCorpus(collection, language, tae.getName(), gis2);
					gis2.close();
					bis.close();
					//Thread.sleep(500);
				}
				num = indexed.get(language) + 1;
				indexed.put(language, num);
				if (num == maxIndexed) {
					break;
				}
			}
			tis.close();
		} catch (IOException e) {
			e.printStackTrace();
		//} catch (InterruptedException e) {
		//	e.printStackTrace();
		}
	}

	private void readCorpus(String collection, Language language, String title, GzipCompressorInputStream gis) {
		SAXBuilder sb = new SAXBuilder();
		try {
			org.jdom2.Document document = sb.build(gis);
			Corpus instance = new Corpus(collection, language, title, "Opus");
			Splitter splitter = ranangSplitter == null || !language.equals(Language.NO) ? getSplitter(language) : ranangSplitter;
			List<String> tokens = readSentences(document.getRootElement().getChildren());
			if (splitter != null) {
				tokens = splitter.process(tokens);
			}
			List<Term> terms = getPreprocessor(language).process(tokens);
			instance.addTerms(terms);
			instance.setId(corpusService.addCorpus(instance));
			System.out.println(instance);
		} catch (JDOMException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private List<String> readSentences(List<Element> sentenceList) {
		List<String> results = new ArrayList<String>();
		for (Element sentence : sentenceList) {
			if ("s".equals(sentence.getName())) {
				results.addAll(readTokens(sentence.getChildren()));
			} else {
				results.addAll(readSentences(sentence.getChildren()));
			}
		}
		return results;
	}

	private List<String> readTokens(List<Element> termList) {
		List<String> results = new ArrayList<String>();
		for (Element token : termList) {
			if ("w".equals(token.getName())) {
				results.add(token.getText());
				showProgression();
			} else {
				readTokens(token.getChildren());
			}
		}
		return results;
	}

	public RawCorpusAccess searchSource(String searchCollection, Language searchLanguage, String searchTitle) {
		for (File collectionDir : sourceFolder.listFiles()) {
			if (!collectionDir.isDirectory() || !collectionDir.getName().equals(searchCollection)) {
				continue;
			}
			return searchCollection(collectionDir, searchLanguage, searchTitle);
		}
		return null;
	}

	private RawCorpusAccess searchCollection(File collectionDir, Language searchLanguage, String searchTitle) {
		String shortName;
		Language language;
		RawCorpusAccess result = null;
		for (File languageFile : collectionDir.listFiles()) {
			if (!languageFile.isFile() || !languageFile.getName().endsWith(".tar.gz")) {
				continue;
			}
			shortName = languageFile.getName();
			shortName = shortName.substring(0, shortName.indexOf(".")).toUpperCase();
			language = Language.valueOf(shortName);
			if (!searchLanguage.equals(language)) {
				continue;
			}
			try {
				FileInputStream fis = new FileInputStream(languageFile);
				GzipCompressorInputStream gis = new GzipCompressorInputStream(fis);
				result = searchTarball(gis, searchTitle);
				gis.close();
				fis.close();
				if (result != null) {
					break;
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return result;
	}

	private RawCorpusAccess searchTarball(GzipCompressorInputStream gis, String searchTitle) {
		TarArchiveInputStream tis = new TarArchiveInputStream(gis);
		try {
			TarArchiveEntry tae;
			while((tae = tis.getNextTarEntry()) != null) {
				if (searchTitle.equals(tae.getName())) {
					byte[] data = new byte[(int) tae.getSize()];
					DataInputStream dis = new DataInputStream(tis);
					dis.read(data);
					ByteArrayInputStream bis = new ByteArrayInputStream(data);
					GzipCompressorInputStream gis2 = new GzipCompressorInputStream(bis);
					InputStreamReader isr = new InputStreamReader(gis2);
					BufferedReader br = new BufferedReader(isr);
					StringBuilder sb = new StringBuilder();
					sb.append(br.readLine() + "\n");
					sb.append("<?xml-stylesheet type=\"text/xsl\" href=\"/corpus.xsl\"?>\n");
					String line;
					while((line = br.readLine()) != null) {
						sb.append(line + "\n");
					}
					br.close();
					isr.close();
					data = sb.toString().getBytes("utf-8");
					return new RawCorpusAccess(new ByteArrayInputStream(data), data.length, "text/xml; charset=utf-8");
				}
			}
			tis.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;
	}

	private Splitter getSplitter(Language language) {
		if (external == null) {
			return null;
		}
		if (splitter == null) {
			splitter = new HashMap<Language, Splitter>();
		}
		if (!splitter.containsKey(language)) {
			splitter.put(language, new StatisticalSplitter(language, external, false));
		}
		return splitter.get(language);
	}

	private Preprocessor getPreprocessor(Language language) {
		if (!preprocessors.containsKey(language)) {
			preprocessors.put(language, new Preprocessor(language));
		}
		return preprocessors.get(language);
	}

	private void showProgression() {
		if (++counter % 100 == 0) {
			System.out.print(counter + "\r");
		}
	}
}
