package impl.indexing.parsing;

import impl.language.Language;
import impl.language.Stopword;
import impl.language.tagging.OsloBergenTagger;
import impl.model.*;
import org.apache.commons.compress.compressors.gzip.GzipCompressorInputStream;

import java.io.*;
import java.util.*;

public class TenTenCorusParser implements CorpusParser {
	private Language language;
	private File sourceFolder;
	private CorpusService corpusService;
	private boolean shouldRun;
	private int counter;
	private int partitionSize;
	private Stopword stopper;

	public TenTenCorusParser(Language language) {
		this.language = language;
		shouldRun = true;
		partitionSize = 1000;
		stopper = new Stopword(language);
	}

	public void setSource(Object source) throws IllegalArgumentException {
		if (source == null || !(source instanceof File) || !((File) source).isDirectory()) {
			throw new IllegalArgumentException("Only directories are supported by the parser.");
		}
		sourceFolder = (File) source;
	}

	public void setCorpusService(CorpusService cs) {
		corpusService = cs;
	}

	public void indexFromSource() throws IllegalStateException {
		if (sourceFolder == null) {
			throw new IllegalStateException("No source is provided.");
		}
		if (corpusService == null) {
			throw new IllegalStateException("CorpusService is not provided.");
		}
		readSource();
	}

	public RawCorpusAccess getRawCorpusAccess(String searchCollection, Language language, String title) throws Exception {
		if (corpusService == null) {
			throw new Exception("CorpusService is not provided.");
		}
		return null;//searchSource(searchCollection, language, title);
	}

	public void terminate() {
		shouldRun = false;
		System.out.println("Terminating TenTen parser...");
	}

	public void readSource() {
		for (File inputFile : sourceFolder.listFiles()) {
			if (!shouldRun) {
				System.out.println("Terminated TenTen parser");
				return;
			}
			if (inputFile.isDirectory() || !inputFile.getName().endsWith(".gz")) {
				continue;
			}
			readInputFile(inputFile);
		}
	}

	private void readInputFile(File inputFile) {
		try {
			FileInputStream fis = new FileInputStream(inputFile);
			GzipCompressorInputStream gis = new GzipCompressorInputStream(fis);
			partitionInput(inputFile.getName(), gis);
			gis.close();
			fis.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void partitionInput(String collection, GzipCompressorInputStream gis) {
		//System.out.println(">>>>>>>>>>" + collection);
		try {
			InputStreamReader isr = new InputStreamReader(gis);
			BufferedReader br = new BufferedReader(isr);
			List<String> partition;
			String line;
			boolean inDoc;
			counter = 0;
			do {
				if (!shouldRun) {
					System.out.println("Terminated TenTen parser");
					return;
				}
				partition = new ArrayList<String>();
				for (int j = 0; j < partitionSize; j++) {
					inDoc = false;
					while ((line = br.readLine()) != null) {
						if (line.replaceFirst("\\<doc\\s+.+\\>", "").equals("")) {
							//System.out.println(line);
							inDoc = true;
						} else if (!inDoc) {
							continue;
						} else if (line.equals("</doc>")) {
							break;
						} else {
							try {
								partition.add(Useful.stripElements(line));
							} catch(StringIndexOutOfBoundsException e) {
								e.printStackTrace();
							}
						}
					}
				}
				if (partition.size() == 0) {
					break;
				}
				savePartition(collection, counter++, partition);
			} while(true);
			br.close();
			isr.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void savePartition(String collection, int partId, List<String> partition) {
		String docId = (partitionSize * partId) + "-" + (partitionSize * (1 + partId));
		if (corpusService.exists(collection, language, docId)) {
			System.err.println("Skipping: " + docId + " " + collection);
			return;
		}
		String[] split;
		List<Term> terms = new ArrayList<Term>();
		for (String line : partition) {
			split = line.split("\t");
			if (split.length < 3) {
				continue;
			}
			//System.out.println(Arrays.toString(split));
			terms.add(new Term(split[1], OsloBergenTagger.convertPos(split[2])));
		}
		terms = stopper.processTerms(terms);
		//for (Term term : terms) {
		//	System.out.println(term);
		//}
		Corpus instance = new Corpus(collection, language, docId, "noTT");
		instance.addTerms(terms);
		instance.setId(corpusService.addCorpus(instance));
		System.out.println(instance);
	}
}
