package impl.language;

import impl.model.*;

import java.util.*;

public class Translator {
	private Database database;

	public Translator(Database database) {
		this.database = database;
	}

	public Set<Query> translate(Query original, Language target, boolean grammar, Pruning pruning) {
		//System.out.println(original + " " + target);
		if (target == null) {
			Set<Query> translations = new HashSet<Query>();
			for (Language version : Language.values()) {
				translations.addAll(translate(original, version, grammar, pruning));
			}
			return translations;
		}
		else if (target.equals(original.getLanguage())) {
			return Collections.singleton(original);
		}
		else if (!original.getLanguage().equals(Language.NO) && !target.equals(Language.NO)) {
			Set<Query> result = new HashSet<Query>();
			for (Query middle : translate(original, Language.NO, grammar, pruning)) {
				result.addAll(translate(middle, target, grammar, pruning));
			}
			return result;
		}
		switch (pruning) {
			case REVERSE:
				return Collections.singleton(reversePrune(original, target, grammar));
			case BALANCE:
				return Collections.singleton(balance(original, target, grammar));
		}
		return null;
	}

	private Query reversePrune(Query original, Language target, boolean grammar) {
		LexiconService ls = new LexiconService(database);
		Lexicon forwardLexicon = ls.getLexicon(original.getLanguage(), target);
		Lexicon reverseLexicon = ls.getLexicon(target, original.getLanguage());
		Query translation = new Query(target);
		List<Term> suggestions;
		List<Term> feedback;
		for (Map.Entry<Term, Double> entry : original.getVector().entrySet()) {
			suggestions = forwardLexicon.getTranslations(entry.getKey());
			feedback = reversePruneTerm(entry.getKey(), suggestions, reverseLexicon, grammar);
			if (grammar && feedback.size() == 0) {
				feedback = reversePruneTerm(entry.getKey(), suggestions, reverseLexicon, false);
			}
			if (feedback.size() == 0) {
				feedback = Arrays.asList(entry.getKey());
			}
			for (Term sugg : feedback) {
				translation.addTerm(sugg);
			}
		}
		return translation;
	}

	private List<Term> reversePruneTerm(Term original, List<Term> suggestions, Lexicon reverse, boolean grammar) {
		if (!grammar) {
			original.pos = null;
		}
		List<Term> feedback = new ArrayList<Term>();
		boolean found;
		for (Term sugg : suggestions) {
			if (!grammar) {
				sugg.pos = null;
			}
			if (!((sugg.pos == null && original.pos == null) || (original.pos != null && original.pos.equals(sugg.pos)))) {
				continue;
			}
			found = false;
			for (Term rev : reverse.getTranslations(sugg)) {
				if (!grammar) {
					rev.pos = null;
				}
				if (!((sugg.pos == null && rev.pos == null) || (sugg.pos != null && sugg.pos.equals(rev.pos)))) {
					continue;
				}
				if (rev.lemma.equals(original.lemma)) {
					found = true;
					continue;
				}
			}
			if (found) {
				feedback.add(sugg);
			}
		}
		return feedback;
	}

	private Query balance(Query original, Language target, boolean grammar) {
		LexiconService ls = new LexiconService(database);
		Lexicon lexicon = ls.getLexicon(original.getLanguage(), target);
		Query translation = new Query(target);
		List<Term> suggestions;
		List<Term> feedback;
		double divide;
		for (Map.Entry<Term, Double> entry : original.getVector().entrySet()) {
			suggestions = lexicon.getTranslations(entry.getKey());
			feedback = balanceTerm(entry.getKey(), suggestions, grammar);
			if (grammar && feedback.size() == 0) {
				feedback = balanceTerm(entry.getKey(), suggestions, false);
			}
			if (feedback.size() == 0) {
				feedback = Arrays.asList(entry.getKey());
			}
			divide = feedback.size();
			for (Term sugg : feedback) {
				translation.addTerm(sugg, (1.0 / divide));
			}
		}
		return translation;
	}

	private List<Term> balanceTerm(Term original, List<Term> suggestions, boolean grammar) {
		if (!grammar) {
			return suggestions;
		}
		List<Term> feedback = new ArrayList<Term>();
		for (Term sugg : suggestions) {
			if ((sugg.pos == null && original.pos == null) || (sugg.pos != null && sugg.pos.equals(original.pos))) {
				feedback.add(sugg);
			}
		}
		return feedback;
	}

	public enum Pruning {
		REVERSE, BALANCE;
	}
}
