package impl.language.splitting;
import impl.evaluation.splitting.SplitOptimiser;
import impl.language.Language;
import impl.model.*;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class RanangSplitter implements Splitter {
	private Lexicon lexicon;
	private Set<Character> consonants;
	//private Map<String, List<String[]>> interpretations;
	private Map<String, String[]> dictionary;

	public RanangSplitter(Lexicon lexicon) {
		this.lexicon = lexicon;
		consonants = new HashSet<Character>(Arrays.asList('b','d','f','g','h','j','k','l','m','n','p','q','r','s','t','v','w','x','z'));
		//interpretations = new HashMap<String, List<String[]>>();
		dictionary = new HashMap<String, String[]>();
	}

	public String[] process(String word) {
		if (word.length() < 2) {
			return new String[] { word };
		}
		//interpretations = new HashMap<String, List<String[]>>();
		System.out.print(word + "\r");
		return getCompoundsSorted(word).get(0).tree.toArray();
	}

	public String[] process(String[] words) {
		List<String> result = process(new ArrayList<String>(Arrays.asList(words)));
		String[] array = new String[result.size()];
		result.toArray(array);
		return array;
	}

	public List<String> process(List<String> words) {
		List<String> output = new ArrayList<String>();
		for (String word : words) {
			if (dictionary.containsKey(word.toLowerCase())) {
				for (String part : dictionary.get(word.toLowerCase())) {
					output.add(part);
				}
				continue;
			}
			dictionary.put(word.toLowerCase(), process(word));
			for (String part : dictionary.get(word.toLowerCase())) {
				output.add(part);
			}
		}
		return output;
	}

	private List<TreeAnalysis> getCompoundsSorted(String word) {
		List<TreeAnalysis> analyses = getCompoundsUnsorted(word);
		Set<Prioritisation> prioritisedAnalyses = prioritiseAnalyses(analyses);
		SortedSet<Prioritisation> sortedAnalysesSet = new TreeSet<Prioritisation>(new Comparator<Prioritisation>() {
			public int compare(Prioritisation o1, Prioritisation o2) {
				int diff;
				for (int i = 0; i < Math.max(o1.heuristics.length, o2.heuristics.length); i++) {
					diff = o1.heuristics[i] - o2.heuristics[i];
					if (diff != 0) {
						return diff;
					}
				}
				return 0;
			}
		});
		sortedAnalysesSet.addAll(prioritisedAnalyses);
		List<TreeAnalysis> sortedAnalyses = new ArrayList<TreeAnalysis>();
		for (Prioritisation entry : sortedAnalysesSet) {
			sortedAnalyses.add(entry.analysis);
		}
		return sortedAnalyses;
	}

	/*
	private List<TreeAnalysis> getCompoundsUnsorted(String word) {
		List<TreeAnalysis> analyses = new ArrayList<TreeAnalysis>();
		for (String[] parts : splitWord(word)) {
			//System.out.println(Arrays.toString(parts));
			for (ParseTreeNode tree : getCompleteWordParses(parts)) {
				analyses.add(analyseParse(tree));
			}
		}
		return analyses;
	}
	*/

	private List<TreeAnalysis> getCompoundsUnsorted(String word) {
		List<TreeAnalysis> analyses = new ArrayList<TreeAnalysis>();
		/*
		for (int k = 0; k < Math.min(7, word.length()); k++) {
			for (String[] parts : split(word, k)) {
				for (ParseTreeNode tree : getCompleteWordParses(parts)) {
					analyses.add(analyseParse(tree));
				}
			}
		}
		*/
		for (String[] parts : splitWord(word)) {
			for (ParseTreeNode tree : getCompleteWordParses(parts)) {
				analyses.add(analyseParse(tree));
			}
		}
		return analyses;
	}

	/*
	private List<String[]> split(String word, int k) {
		if (k == 0 || word.length() == 1) {
			return Collections.singletonList(new String[]{word});
		}
		List<String[]> splits = new ArrayList<String[]>();
		String[] newArr1;
		//String[] newArr2;
		for (int i = 1; i < word.length(); i++) {
			for (String[] tail : split(word.substring(i), k - 1)) {
				newArr1 = new String[tail.length + 1];
				//newArr2 = new String[tail.length + 1];
				newArr1[0] = word.substring(0, i);
				//newArr2[0] = stemmer.optimisationProcess(word.substring(0, i));
				int j = 0;
				for (String t : tail) {
					newArr1[++j] = t;
					//newArr2[j] = t;
				}
				splits.add(newArr1);
				//splits.add(newArr2);
			}
		}
		return splits;
	}
	*/

	private List<String[]> splitWord(String word) {
		return splitWord(word, new HashMap<String, List<String[]>>(), 7);
	}

	private List<String[]> splitWord(String word, Map<String, List<String[]>> interpretations, int k) {
		if (interpretations.containsKey(word)) {
			return interpretations.get(word);
		}
		try {
			interpretations.put(word, new ArrayList<String[]>());
			interpretations.get(word).add(new String[]{word});
			if (k == 0) {
				return interpretations.get(word);
			}
			String[] newArr;
			newArr = new String[1];
			newArr[0] = word;
			if (getPartOfWordParses(newArr).size() > 0) {
				interpretations.get(word).add(newArr);
			}
			String[] split;
			loop:
			for (int i = word.length() - 1; i > 0; i--) {
				split = new String[]{word.substring(0, i), word.substring(i)};
				for (String[] firstLast : combineParts(split[0], split[1])) {
					newArr = new String[1];
					newArr[0] = firstLast[1];
					if (getPartOfWordParses(newArr).size() == 0) {
						continue loop;
					}
					for (String[] suggestedStart : splitWord(firstLast[0], interpretations, k - 1)) {
						newArr = new String[suggestedStart.length + 1];
						System.arraycopy(suggestedStart, 0, newArr, 0, suggestedStart.length);
						newArr[suggestedStart.length] = firstLast[1];
						interpretations.get(word).add(newArr);
					}
				}
			}
			return interpretations.get(word);
		} catch (OutOfMemoryError e) {
			e.printStackTrace();
			return Arrays.asList(new String[][]{ { word } });
		}
	}

	private List<ParseTreeNode> getCompleteWordParses(String[] parts) {
		//if (parts.length > 1) {
		//	System.out.println(parts);
		//}
		if (isAssumedCruft(parts)) {
			return new ArrayList<ParseTreeNode>();
		}
		List<ParseTreeNode> filteredTrees = new ArrayList<ParseTreeNode>();
		for (ParseTreeNode tree : parse(parts)) {
			if (isCompleteWord(tree)) {
				//System.out.println("comp: " + tree);
				filteredTrees.add(tree);
			}
		}
		return filteredTrees;
	}

	private List<ParseTreeNode> getPartOfWordParses(String[] parts) {
		if (isAssumedCruft(parts)) {
			return new ArrayList<ParseTreeNode>();
		}
		List<ParseTreeNode> filteredTrees = new ArrayList<ParseTreeNode>();
		for (ParseTreeNode tree : parse(parts)) {
			if (isPartOfWord(tree)) {
				//System.out.println("part: " + tree);
				filteredTrees.add(tree);
			}
		}
		return filteredTrees;
	}

	private boolean isCompleteWord(ParseTreeNode tree) {
		for (String part : tree.toArray()) {
			if (part.length() == 1 && !"s".equals(part) && !"e".equals(part)) {
				return false;
			}
		}
		return true;
		/*
		String[] seq = tree.toArray();
		int i = 0;
		for (String part : seq) {
			if (++i == 1 && (part.equals("s") || part.equals("e")) && i < seq.length) {
				continue;
			}
			if (!lexicon.hasWord(part)) {
				return false;
			}
		}
		return true;
		*/
	}

	private boolean isPartOfWord(ParseTreeNode tree) {
		for (String part : tree.toArray()) {
			if ("s".equals(part) || "e".equals(part)) {
				return false;
			}
		}
		return true;
	}

	private Set<ParseTreeNode> parse(String[] parts) {
		if (parts.length == 1) {
			return Collections.singleton((ParseTreeNode) new LeafParseTreeNode(parts[0]));
		}
		Set<ParseTreeNode> trees = new HashSet<ParseTreeNode>();
		String[] left;
		String[] right;
		for (int i = 1; i < parts.length; i++) {
			left = new String[i];
			right = new String[parts.length - i];
			System.arraycopy(parts, 0, left, 0, left.length);
			System.arraycopy(parts, left.length, right, 0, right.length);
			for (ParseTreeNode leftTree : parse(left)) {
				for (ParseTreeNode rightTree : parse(right)) {
					trees.add(new InnerParseTreeNode(leftTree, rightTree));
				}
			}
		}
		return trees;
	}

	private boolean isAssumedCruft(String[] parts) {
		int num = parts.length;
		int sum = 0;
		for (String cand : parts) {
			sum += cand.length();
		}
		return parts.length > 2 && (sum / num) < 3;
	}

	private String[][] combineParts(String a, String b) {
		int length = a.length();
		if (a.length() >= 2 && consonants.contains(a.charAt(length - 1)) && a.charAt(length - 2) == a.charAt(length - 1)) {
			return new String[][]{ { a, b }, { String.valueOf(a.charAt(length - 1)) , b } };
		}
		else {
			return new String[][]{ { a, b } };
		}
	}

	private Set<Prioritisation> prioritiseAnalyses(List<TreeAnalysis> analyses) {
		Set<Prioritisation> prioritisations = new HashSet<Prioritisation>();
		for (TreeAnalysis analysis : analyses) {
			prioritisations.add(prioritise(analysis, analyses));
		}
		return prioritisations;
	}

	private Prioritisation prioritise(TreeAnalysis analysis, List<TreeAnalysis> analyses) {
		return new Prioritisation(analysis, new int[]{
				analysis.getNumberOfCompounds(),
				- (analysis.hasPreferableEpentheticS() ? 1 : 0),
				- (analysis.hasAdjectiveNounAmbiguity(analyses) ? 1 : 0),
				analysis.getPosWeightMapping()[0],
				analysis.hasSuffixHead() ? 1 : 0,
				- (analysis.firstMemberIsCompound() ? 1 : 0),
				- (analysis.hasAmbiguousEpentheticE(analyses) ? 1 : 0),
				analysis.containsInflectionFormatives() ? 1 : 0,
				analysis.containsEpentheses() ? 1 : 0,
				- (analysis.firstMemberIsProbablyCompound(analyses) ? 1 : 0),
				analysis.getReversedNegatedPartLengths(),
				analysis.firstMemberIsVerbal() ? 1 : 0,
				analysis.getPosWeightMappingAdjusted()
		});
	}

	private TreeAnalysis analyseParse(ParseTreeNode tree) {
		return new TreeAnalysis(tree);
	}

	private static class Prioritisation {
		public TreeAnalysis analysis;
		public int[] heuristics;

		public Prioritisation(TreeAnalysis analysis, int[] heuristics) {
			this.analysis = analysis;
			this.heuristics = heuristics;
		}
	}

	private static class TreeAnalysis {
		public ParseTreeNode tree;

		public TreeAnalysis(ParseTreeNode tree) {
			this.tree = tree;
		}

		public int getNumberOfCompounds() {
			int number = 0;
			for (String part : tree.toArray()) {
				if (!"s".equals(part) && !"e".equals(part)) {
					number++;
				}
			}
			return number;
		}

		public boolean hasPreferableEpentheticS() {
			Pattern p = Pattern.compile("/\\S+(ed|ng)\\ss\\s/i");
			Matcher m = p.matcher(tree.toString());
			return m.find();
		}

		public boolean hasAdjectiveNounAmbiguity(List<TreeAnalysis> analyses) {
			return false; // TODO
		}

		public int[] getPosWeightMapping() {
			return new int[1]; // TODO
		}

		public boolean hasSuffixHead() {
			return false; // TODO
		}

		public boolean firstMemberIsCompound() {
			Pattern p = Pattern.compile("/^(\\S+\\s){2,}s(\\s|$)/i");
			Matcher m = p.matcher(tree.toString());
			return m.find();
		}

		public boolean hasAmbiguousEpentheticE(List<TreeAnalysis> analyses) {
			return false; // TODO
		}

		public boolean containsInflectionFormatives() {
			return false; // TODO
		}

		public boolean containsEpentheses() {
			String val = tree.toString();
			return val.toLowerCase().indexOf(" s ") != -1 || val.toLowerCase().indexOf(" e ") != -1;
		}

		public boolean firstMemberIsProbablyCompound(List<TreeAnalysis> analyses) {
			return false; // TODO
		}

		public int getReversedNegatedPartLengths() {
			return 0; // TODO
		}

		public boolean firstMemberIsVerbal() {
			return false; // TODO
		}

		public int getPosWeightMappingAdjusted() {
			return 0; // TODO
		}
	}

	private interface ParseTreeNode {
		@Override
		String toString();
		String[] toArray();
	}

	private static class LeafParseTreeNode implements ParseTreeNode {
		public String value;

		public LeafParseTreeNode(String value) {
			this.value = value;
		}

		public String toString() {
			return value;
		}

		public String[] toArray() {
			return new String[]{ value };
		}
	}

	private static class InnerParseTreeNode implements ParseTreeNode {
		public ParseTreeNode left;
		public ParseTreeNode right;

		public InnerParseTreeNode(ParseTreeNode left, ParseTreeNode right) {
			this.left = left;
			this.right = right;
		}

		public String toString() {
			return left + " " + right;
		}

		public String[] toArray() {
			String[] leftArr = left.toArray();
			String[] rightArr = right.toArray();
			String[] resArr = new String[leftArr.length + rightArr.length];
			System.arraycopy(leftArr, 0, resArr, 0, leftArr.length);
			System.arraycopy(rightArr, 0, resArr, leftArr.length, rightArr.length);
			return resArr;
		}
	}

	public static void main(String[] args) {
		Database db = new Database(Environment.getInstance().getProperty("database"));
		String[] cases = new String[]{
				"bil",
				"utenomjordisk",
				"tungvektsløfter",
				"musikkspiller",
				"skolegårdsringeklokke",
				"billedkunstner",
				"askebeger",
				"rulleskøyte",
				"gårdsplass",
				"rutebilstasjon",
				"hverdagsrutine",
				"frivillighetsarbeid",
				"kinarestaurant",
				"bursdagsbarn",
				"universitet",
				"usikker",
				"oppmerksomhet",
				"innbille",
				"innrømme",
				"godta",
				"hverandre",
				"unødvendig",
				"middagsgjest",
		};
		LexiconService ls = new LexiconService(db);
		RanangSplitter rs = new RanangSplitter(ls.getLexicon(Language.NO));
		//List<TreeAnalysis> tal;
		//for (String test : cases) {
		//	tal = rs.getCompoundsSorted(test);
		//	System.out.println(test + ":");
		//	for (TreeAnalysis ta : tal) {
		//		print(ta.tree, 0);
		//	}
		//}
		eval(rs, SplitOptimiser.getNorwegianFinalTest());
	}

	private static void print(ParseTreeNode tree, int i) {
		for (int j = 0; j < i; j++) {
			System.out.print("\t");
		}
		if (tree instanceof LeafParseTreeNode) {
			System.out.println("* " + ((LeafParseTreeNode) tree).value);
		}
		else {
			System.out.println("*");
			print(((InnerParseTreeNode) tree).left, i + 1);
			print(((InnerParseTreeNode) tree).right, i + 1);
		}
	}

	private static void eval(RanangSplitter rs, Map<String, String[]> test) {
		int right = 0, wrong = 0, tp = 0, fp = 0, tn = 0, fn = 0;
		double p, r;
		List<TreeAnalysis> tals;
		TreeAnalysis tal;
		for (Map.Entry<String, String[]> entry : test.entrySet()) {
			System.out.println(entry.getKey() + " " + Arrays.toString(entry.getValue()));
			tals = rs.getCompoundsSorted(entry.getKey());
			for (TreeAnalysis t : tals) {
				print(t.tree, 0);
			}
			tal = tals.get(0);
			if (Arrays.deepEquals(test.get(entry.getKey()), tal.tree.toArray())) {
				// tp or tn
				right++;
				if (entry.getKey().equals(test.get(entry.getKey())[0])) {
					tn++;
				} else {
					tp++;
				}
			} else {
				// fn or fp
				wrong++;
				if (entry.getKey().equals(test.get(entry.getKey())[0])) {
					fn++;
				} else {
					fp++;
				}
			}
		}
		//System.out.println(tp + " " + fp + " " + fn + " " + tn);
		p = tp == .0 && fp == .0 ? .0 : tp / (tp + fp);
		r = tp == .0 && fn == .0 ? .0 : tp / (tp + fn);
		//System.out.println(p + " " + r);
		double[] s = new double[]{
				2.0 * ((p * r) / (p + r)),
				(double) right / (double) (right + wrong)
		};
		System.out.println("f1=" + s[0] + ",accuracy=" + s[1]);
	}
}
