package impl.language.splitting;

import impl.language.Language;
import impl.model.*;

import java.io.*;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

public class StatisticalSplitter extends AbstractSplitter implements Splitter {
	private Language language;
	private Database database;
	private Lexicon lexicon;
	//private Map<String, Set<String[]>> interpretations;
	private Map<String, Integer> frequencies;
	private Map<String, String[]> dictionary;
	private boolean incognito;
	private long lastSave;
	private Map<String, String[]> saveList;
	private double penalty;
	private double averageLength;
	private List<MorphologicalOperation> globalMops;

	public StatisticalSplitter(Language language, Database database, boolean incognito) {
		super();
		this.language = language;
		this.database = database;
		LexiconService ls = new LexiconService(database);
		this.lexicon = ls.getLexicon(language);
		//interpretations = new HashMap<String, Set<String[]>>();
		frequencies = new HashMap<String, Integer>();
		this.incognito = incognito;
		if (!incognito) {
			load();
		}
		penalty = 9.688;
		calculateAverageLength();
		globalMops = getGlobalMops();
		}

	public void setPenalty(double penalty) {
		this.penalty = penalty;
	}

	public String[] process(String word) {
		List<String> result = process(Arrays.asList(word));
		String[] array = new String[result.size()];
		result.toArray(array);
		return array;
	}

	public String[] process(String[] words) {
		List<String> result = process(new ArrayList<String>(Arrays.asList(words)), false);
		String[] array = new String[result.size()];
		result.toArray(array);
		return array;
	}

	public List<String> process(List<String> words) {
		return language.equals(Language.NO) ? process(words, false) : words;
	}

	public List<String> process(List<String> words, boolean print) {
		Map<String, String[]> replacements = new HashMap<String, String[]>();
		for (String word : words) {
			replacements.put(word.toLowerCase(), null);
		}
		Set<String> unprocessed = new HashSet<String>();
		for (Map.Entry<String, String[]> entry : replacements.entrySet()) {
			if (!incognito && dictionary.containsKey(entry.getKey())) {
				replacements.put(entry.getKey(), dictionary.get(entry.getKey()) == null ? new String[]{entry.getKey()} : dictionary.get(entry.getKey()));
			} else {
				unprocessed.add(entry.getKey());
			}
		}
		for (Map.Entry<String, String[]> entry : getReplacements(unprocessed, print).entrySet()) {
			System.out.println(entry.getKey() + ": " +  Arrays.toString(entry.getValue()));
			replacements.put(entry.getKey(), entry.getValue());
			try {
				dictionary.put(entry.getKey(), entry.getKey().equals(entry.getValue()[0]) ? null : entry.getValue());
			} catch (NullPointerException e) {
				e.printStackTrace();
			}
			if (!incognito) {
				save(entry.getKey(), entry.getValue());
			}
		}
		List<String> output = new ArrayList<String>();
		for (int i = 0; i < words.size(); i++) {
			//System.out.println(words.get(i));
			try {
				for (String replace : replacements.get(words.get(i).toLowerCase())) {
					//System.out.println("\t" + replace);
					output.add(replace);
				}
			} catch (NullPointerException e) {
				System.err.println(words.get(i));
			}
		}
		return output;
	}

	public Map<String, String[]> getReplacements(Set<String> words, boolean print) {
		Map<String, Set<String[]>> unsortedCandidates = getCandidatesUnsorted(words);
		Map<String, Set<String[]>> filteredCandidates = new HashMap<String, Set<String[]>>();
		Set<String[]> allCandidates = new HashSet<String[]>();
		Set<String[]> approvedCandidates;
		Map<String, String[]> replacements = new HashMap<String, String[]>();
		for (Map.Entry<String, Set<String[]>> entry : unsortedCandidates.entrySet()) {
			approvedCandidates = new HashSet<String[]>();
			loop:
			for (String[] cand : entry.getValue()) {
				cand = applyMorphologicalOperations(cand);
				//if (print) {
				//	System.out.println(Arrays.toString(cand));
				//}
				approvedCandidates.add(cand);
				allCandidates.add(cand);
			}
			if (approvedCandidates.size() > 0) {
				filteredCandidates.put(entry.getKey(), approvedCandidates);
			}
		}
		//System.out.println("test1");
		retrieveFrequencies(allCandidates);
		//System.out.println("test2");
		List<String[]> sortedCandidates;
		String[] best;
		//int i;
		for (Map.Entry<String, Set<String[]>> entry : filteredCandidates.entrySet()) {
			sortedCandidates = sortCandidates(entry.getValue());
			best = sortedCandidates.get(0);
			replacements.put(entry.getKey(), best);
			if (print) {
				/*
				i = 0;
				for (String[] cand : sortedCandidates) {
					System.out.println(Arrays.toString(cand));
					if (++i == 10) {
						break;
					}
				}
				*/
				System.out.println(Arrays.toString(best) + " " + score(best));
			}
		}
		return replacements;
	}

	private Map<String, Set<String[]>> getCandidatesUnsorted(Set<String> words) {
		int length = words.size();
		int i = 0;
		int maxThreads = 250;
		final Map<String, Set<String[]>> candidates = new HashMap<String, Set<String[]>>();
		if (length > maxThreads) {
			Set<String> chunk = new HashSet<String>();
			for (String word : words) {
				chunk.add(word);
				if (++i % maxThreads == 0) {
					candidates.putAll(getCandidatesUnsorted(chunk));
					chunk = new HashSet<String>();
				}
			}
			if (i % maxThreads != 0) {
				candidates.putAll(getCandidatesUnsorted(chunk));
			}
			return candidates;
		}
		Thread[] threads = new Thread[length];
		for (final String word : words) {
			threads[i] = new Thread() {
				@Override
				public void run() {
					candidates.put(word, getCandidatesUnsorted(word));
				}
			};
			threads[i++].start();
		}
		for (int j = 0; j < length; j++) {
			try {
				threads[j].join();
				threads[j] = null;
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		return candidates;
	}

	private String[] applyMorphologicalOperations(String[] cand) {
		if (cand.length <= 2) {
			return cand;
		}
		List<String> pure = new ArrayList<String>();
		pure.add(cand[0]);
		loop:
		for (int i = 1; i < cand.length - 1; i++) {
			for (MorphologicalOperation mop : globalMops) {
				if (cand[i].equals(mop.decompounded)) {
					continue loop;
				}
			}
			pure.add(cand[i]);
		}
		pure.add(cand[cand.length - 1]);
		cand = new String[pure.size()];
		pure.toArray(cand);
		return cand;
	}

	public Map<String, String[]> optimisationProcess(Set<String> words, final boolean print) {
		Map<String, Set<String[]>> unsortedCandidates = getCandidatesUnsorted(words);
		Set<String[]> allCandidates = new HashSet<String[]>();
		for (Set<String[]> cand : unsortedCandidates.values()) {
			allCandidates.addAll(cand);
		}
		retrieveFrequencies(allCandidates);
		Map<String, String[]> candidates = new HashMap<String, String[]>();
		String[] best;
		for (Map.Entry<String, Set<String[]>> entry : unsortedCandidates.entrySet()) {
			best = sortCandidates(entry.getValue()).get(0);
			candidates.put(entry.getKey(), best);
			if (print) {
				System.out.println(Arrays.toString(best) + " " + score(best));
			}
		}
		return candidates;
	}

	public Map<MorphologicalOperation, GlobalScore> trainProcess(Set<String> words, final boolean print) {
		Map<String, String[]> decompounds = optimisationProcess(words, false);
		int i = 0;
		final Map<MorphologicalOperation, GlobalScore> mops = new HashMap<MorphologicalOperation, GlobalScore>();
		Thread[] threads = new Thread[decompounds.size()];
		for (final Map.Entry<String, String[]> decompound : decompounds.entrySet()) {
			threads[i] = new Thread() {
				@Override
				public void run() {
					//System.out.println(decompound.getKey() + " " + Arrays.toString(decompound.getValue()));
					for (Map.Entry<MorphologicalOperation, GlobalScore> entry : getMorphologicalOperationHeuristics(decompound.getKey(), decompound.getValue()).entrySet()) {
						if (print) {
							System.out.println(Arrays.toString(decompound.getValue()) + " -> " + entry.getKey() + " " + entry.getValue());
						}
						mops.put(entry.getKey(), mops.containsKey(entry.getKey()) ? entry.getValue().add(mops.get(entry.getKey())) : entry.getValue());
					}
				}
			};
			threads[i++].start();
		}
		for (int j = 0; j < threads.length; j++) {
			try {
				threads[j].join();
				threads[j] = null;
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		return mops;
	}

	private void load() {
		dictionary = new HashMap<String, String[]>();
		lastSave = System.currentTimeMillis();
		saveList = new HashMap<String, String[]>();
		try {
			File file = new File(Environment.getInstance().getProperty("test-cws." + language.toString()));
			if (!file.exists()) {
				file.createNewFile();
			}
			FileReader fr = new FileReader(file);
			BufferedReader br = new BufferedReader(fr);
			String line;
			int pos;
			while ((line = br.readLine()) != null) {
				pos = line.indexOf(">");
				if (pos != -1) {
					dictionary.put(line.substring(0, pos), line.substring(pos + 1).split(" "));
				}
				else {
					dictionary.put(line, null);
				}
			}
			br.close();
			fr.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void save(String word, String[] splits) {
		if (System.currentTimeMillis() - lastSave < 300000L) {
			saveList.put(word, splits.length == 1 && splits[0].equals(word) ? null : splits);
			return;
		}
		flush();
	}

	public void flush() {
		try {
			File out = new File(Environment.getInstance().getProperty("test-cws." + language.toString()));
			if (!out.exists()) {
				out.createNewFile();
			}
			PrintWriter pw = new PrintWriter(new FileOutputStream(out, true));
			for (Map.Entry<String, String[]> entry : saveList.entrySet()) {
				if (entry.getValue() == null) {
					pw.append(entry.getKey() + "\n");
				}
				else {
					pw.append(entry.getKey() + ">" + Useful.join(entry.getValue(), " ") + "\n");
				}
			}
			pw.close();
			dictionary.putAll(saveList);
			saveList = new HashMap<String, String[]>();
			lastSave = System.currentTimeMillis();
		} catch (FileNotFoundException e) {
			//e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private List<String[]> sortCandidates(Set<String[]> candidates) {
		SortedSet<String[]> sortedPartsSet = new TreeSet<String[]>(new Comparator<String[]>() {
			public int compare(String[] o1, String[] o2) {
				int comp = (int) (score(o2) - score(o1));
				if (comp != 0) {
					return comp;
				}
				comp = o2.length - o1.length;
				if (comp != 0) {
					return comp;
				}
				int hs1 = 0;
				int hs2 = 0;
				for (String s : o1) {
					hs1 += s.hashCode();
				}
				for (String s : o2) {
					hs2 += s.hashCode();
				}
				return hs2 - hs1;
			}
		});
		sortedPartsSet.addAll(candidates);
		return new ArrayList<String[]>(sortedPartsSet);
	}

	private Set<String[]> getCandidatesUnsorted(String word) {
		Set<String[]> parts = new HashSet<String[]>();
		for (int k = 0; k < Math.min(7, word.length()); k++) {
			//parts.addAll(split(word, k));
			for (String[] seq : split(word, k)) {
				if (!isAssumedCruft(seq)) {
					parts.add(seq);
				}
			}
		}
		return parts;
	}

	private boolean isAssumedCruft(String[] parts) {
		double num = parts.length;
		double sum = 0;
		for (String cand : parts) {
			sum += cand.length();
		}
		//return false;
		return parts.length > 2 && Math.ceil(sum / num) < Math.floor(Math.sqrt(averageLength));
	}

	private Set<String[]> split(String word, int k) {
		if (k == 0 || word.length() == 1) {
			return Collections.singleton(new String[]{ word });
		}
		Set<String[]> splits = new HashSet<String[]>();
		String[] newArr1;
		//String[] newArr2;
		for (int i = 1; i < word.length(); i++) {
			for (String[] tail : split(word.substring(i), k - 1)) {
				newArr1 = new String[tail.length + 1];
				//newArr2 = new String[tail.length + 1];
				newArr1[0] = word.substring(0, i);
				//newArr2[0] = stemmer.optimisationProcess(word.substring(0, i));
				int j = 0;
				for (String t : tail) {
					newArr1[++j] = t;
					//newArr2[j] = t;
				}
				splits.add(newArr1);
				//splits.add(newArr2);
			}
		}
		return splits;
	}

	private double score(String[] parts) {
		double score = .0;
		//for (int freq : freq(parts)) {
		//	score += Math.log(1 + freq);
		//}
		for (String part : parts) {
			score += Math.log(.0000000000001 + freq(part)) - Math.log(Math.E + .25 * Math.abs((double) part.length() - averageLength));
			//score += Math.log(.0000000000001 + freq(part));
		}
		return score - penalty * (parts.length - 1);
	}

	private void retrieveFrequencies(Set<String[]> candidates) {
		Set<String> parts = new HashSet<String>();
		for (String[] cand : candidates) {
			for (String word : cand) {
				//if (word.length() < 2) {
				//	frequencies.put(word.toLowerCase(), 0);
				//}
				//else {
					parts.add(word.toLowerCase());
				//}
			}
		}
		parts.removeAll(frequencies.keySet());
		if (parts.size() == 0) {
			return;
		}
		try {
			String sql = "SELECT t.lemma AS lemma,SUM(ct.count) AS count FROM term AS t,corpus_term AS ct WHERE " +
					"t.id=ct.term_id AND t.language=? AND lower(t.lemma) IN(" + makePlaceholders(parts.size()) + ") GROUP BY t.lemma;";
			//System.out.println(sql);
			PreparedStatement ps = database.getConnection().prepareStatement(sql);
			int i = 1;
			ps.setString(i, language.toString());
			for (String word : parts) {
				//System.out.println(word);
				ps.setString(++i, word);
			}
			ResultSet rs = ps.executeQuery();
			//System.out.println(parts.size());
			while (rs.next()) {
				//if (rs.getString("lemma") == null) {
				//	continue;
				//}
				frequencies.put(rs.getString("lemma").toLowerCase(), rs.getInt("count"));
				//System.out.println(rs.getString("lemma").toUpperCase() + ": " + rs.getInt("count"));
			}
			rs.close();
		} catch (SQLException e) {
			e.printStackTrace();
		}
	}

	private void calculateAverageLength() {
		try {
			String sql = "SELECT AVG(LENGTH(lemma)) AS length FROM term WHERE language=?;";
			PreparedStatement ps = database.getConnection().prepareStatement(sql);
			int i = 1;
			ps.setString(i, language.toString());
			ResultSet rs = ps.executeQuery();
			if (!rs.next()) {
				return;
			}
			averageLength = rs.getDouble(1);
			System.out.println("average_length=" + averageLength);
			rs.close();
		} catch (SQLException e) {
			e.printStackTrace();
		}
	}

	private int freq(String word) {
		String upper = word.toLowerCase();
		return frequencies.containsKey(upper) ? frequencies.get(upper) : 0;
	}

	private String makePlaceholders(int len) {
		if (len < 1) {
			throw new RuntimeException("No tokens");
		}
		StringBuilder sb = new StringBuilder(len * 2 - 1);
		sb.append("?");
		for (int i = 1; i < len; i++) {
			sb.append(",?");
		}
		return sb.toString();
	}

	public Map<MorphologicalOperation, GlobalScore> getMorphologicalOperationHeuristics(String compounded, String[] compounds) {
		int i = 0;
		boolean[] bindings = new boolean[compounds.length];
		Map<MorphologicalOperation, GlobalScore> mops = new HashMap<MorphologicalOperation, GlobalScore>();
		double completeScore = score(compounds);
		String[] excluded;
		for (String part : compounds) {
			bindings[i++] = !lexicon.hasWord(part);
		}
		//System.out.println("one " + Arrays.toString(compounds) + " " + Arrays.toString(bindings));
		if (i < 2 || bindings[0] || bindings[compounds.length - 1]) {
			//System.out.println(i + " " + bindings[0] + " " + bindings[compounds.length - 1]);
			return mops;
		}
		//System.out.println("two " + Arrays.toString(compounds));
		for (i = 1; i < compounds.length - 1; i++) {
			if (bindings[i] && bindings[i + 1]) {
				return mops;
			}
		}
		//System.out.println("three " + i + " " + Arrays.toString(compounds));
		MorphologicalOperation mo;
		GlobalScore gs;
		for (i = 1; i < compounds.length - 1; i++) {
			excluded = excludeCompound(compounds, i);
			gs = new GlobalScore(1, score(excluded) - completeScore);
			mo = getMorphologicalOperation(compounded, excluded);
			mops.put(mo, mops.containsKey(mo) ? gs.add(mops.get(mo)) : gs);
		}
		return mops;
	}

	private String[] excludeCompound(String[] compounds, int index) {
		String[] shorter = new String[compounds.length - 1];
		System.arraycopy(compounds, 0, shorter, 0, index);
		System.arraycopy(compounds, index+1, shorter, index, compounds.length - index - 1);
		return shorter;
	}

	public MorphologicalOperation getMorphologicalOperation(String compounded, String[] compounds) {
		String decompounded = Useful.join(compounds, ""); // TODO: Lemmatise
		if ("".equals(decompounded)) {
			return new MorphologicalOperation(null, compounded);
		}
		if (compounded.equals(decompounded)) {
			return new MorphologicalOperation(null, null);
		}
		// http://introcs.cs.princeton.edu/java/96optimization/Diff.java.html
		//System.out.println(compounded + " " + decompounded + " " + Arrays.toString(compounds));
		int M = decompounded.length();
		int N = compounded.length();
		int[][] opt = new int[M + 1][N + 1];
		for (int i = M-1; i >= 0; i--) {
			for (int j = N-1; j >= 0; j--) {
				if (decompounded.charAt(i) == compounded.charAt(j)) {
					opt[i][j] = opt[i + 1][j + 1] + 1;
				}
				else {
					opt[i][j] = Math.max(opt[i + 1][j], opt[i][j + 1]);
				}
			}
		}
		String x = "", y = "";
		int i = 0, j = 0;
		while(i < M && j < N) {
			if (decompounded.charAt(i) == compounded.charAt(j)) {
				i++;
				j++;
			}
			else if (opt[i + 1][j] >= opt[i][j + 1]) {
				y += decompounded.charAt(i++);
			}
			else {
				x += compounded.charAt(j++);
			}
		}
		while(i < M || j < N) {
			if (i < M) {
				x += decompounded.charAt(i++);
			}
			else if (j < N) {
				y += compounded.charAt(j++);
			}
		}
		return new MorphologicalOperation("".equals(x) ? null : x, "".equals(y) ? null : y);
	}

	public class MorphologicalOperation {
		public String decompounded;
		public String compounded;

		public MorphologicalOperation(String decompounded, String compounded) {
			this.decompounded = decompounded;
			this.compounded = compounded;
		}

		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof MorphologicalOperation)) {
				return false;
			}
			if (obj == this) {
				return true;
			}
			MorphologicalOperation other = ((MorphologicalOperation) obj);
			return (decompounded == null ? decompounded == other.decompounded : decompounded.equals(other.decompounded))
					&& (compounded == null ? compounded == other.compounded : compounded.equals(other.compounded));
		}

		@Override
		public int hashCode() {
			return ((compounded == null ? 0 : compounded.hashCode()) % 53 + 3) * ((decompounded == null ? 0 : decompounded.hashCode()) % 82 + 43);
		}

		@Override
		public String toString() {
			return "morph(" + decompounded + "/" + compounded + ")";
		}
	}

	public class GlobalScore {
		public int frequency;
		public double improvement;

		public GlobalScore(int frequency, double improvement) {
			this.frequency = frequency;
			this.improvement = improvement;
		}

		public GlobalScore add(GlobalScore gs) {
			this.frequency += gs.frequency;
			this.improvement += gs.improvement;
			return this;
		}

		public double getAverage() {
			return improvement / (double) frequency;
		}

		@Override
		public String toString() {
			return frequency + " " + improvement + " " + getAverage();
		}
	}

	public List<MorphologicalOperation> getGlobalMops() {
		List<MorphologicalOperation> mops = new ArrayList<MorphologicalOperation>();
		try {
			File out = new File(Environment.getInstance().getProperty("test-morph." + language.toString()));
			FileReader fr = new FileReader(out);
			BufferedReader br = new BufferedReader(fr);
			String line = br.readLine();
			String[] split;
			int i = 0;
			while ((line = br.readLine()) != null) {
				if (i != 0 && i != 4) {
					i++;
					continue;
				}
				split = line.replaceFirst("\\/", " ").split("\\s+");
				mops.add(new MorphologicalOperation(split[0], split[1]));//, new GlobalScore(Integer.parseInt(split[2]), Double.parseDouble(split[3])));
				if (++i == 10) {
					break;
				}
			}
			br.close();
			fr.close();
			System.out.println(mops.size() + " " + mops);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return mops;
	}

	public static void main(String[] args) {
		Database database = new Database(Environment.getInstance().getProperty("test-db-ext"));
		String[] cases = new String[]{
				"bil",
				"askebeger",
				"rulleskøyte",
				"billedkunstner",
				"gårdsplass",
				"rutebilstasjon",
				"utenomjordisk",
				"tungvektsløfter",
				"musikkspiller",
				"skolegårdsringeklokke",
				"hverdagsrutine",
				"frivillighetsarbeid",
				"kinarestaurant",
				"bursdagsbarn",
				"universitet",
				"usikker",
				"oppmerksomhet",
				"innbille",
				"innrømme",
				"godta",
				"hverandre",
				"unødvendig",
				"middagsgjest",
		};
		StatisticalSplitter ms = new StatisticalSplitter(Language.NO, database, true);
		String[] cand;
		for (String test : cases) {
			cand = ms.process(test);
			System.out.println(test + ": " + Arrays.toString(cand) + " " + ms.score(cand));
		}
		ms = new StatisticalSplitter(Language.EN, database, false);
		cases = new String[]{
				"flowerpot",
		};
		for (String test : cases) {
			System.out.println(test + ":");
			cand = ms.process(test);
			System.out.println(Arrays.toString(cand) + ms.score(cand));
		}
	}
}
