package impl.model;

import impl.language.Language;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

public class VectorSpace {
	private Database database;
	private Map<Language, Map<Term, Double>> dfs;

	public VectorSpace(Database database) {
		this.database = database;
		dfs = new HashMap<Language, Map<Term, Double>>();
	}

	public double df(Term term, Language language, boolean grammar) {
		if (!dfs.containsKey(language)) {
			dfs.put(language, new HashMap<Term, Double>());
		}
		if (!dfs.get(language).containsKey(term)) {
			dfFromDb(term, language, grammar);
		}
		if (!grammar) {
			return dfs.get(language).get(term);
		}
		double df = .0;
		for (Map.Entry<Term, Double> entry : dfs.get(language).entrySet()) {
			if (term.lemma.equals(entry.getKey().lemma)) {
				df += entry.getValue();
			}
		}
		return df;
	}

	public double idf(Term term, Language language, boolean grammar) {
		double df = df(term, language, grammar);
		if (df == .0) {
			return 1.0;
		}
		return 1.0 - Math.log(df);
	}

	public double tf(Term term, Document document, boolean grammar) {
		if (!grammar) {
			return document.getNormalisedVector().containsKey(term) ? document.getNormalisedVector().get(term) : 0;
		}
		double tf = .0;
		for (Map.Entry<Term, Double> entry : document.getNormalisedVector().entrySet()) {
			if (term.lemma.equals(entry.getKey().lemma)) {
				tf += entry.getValue();
			}
		}
		return tf;
	}

	public double tfIdf(Term term, Document document, boolean grammar) {
//		System.out.println(tf(term, document) + " " + df(term, document.getLanguage()) + " " + idf(term, document.getLanguage()) + " " + term + " " + document);
//		if (document.getLanguage().equals(Language.NO)) {
//			System.out.println("test");
//		}
		return tf(term, document, grammar) * idf(term, document.getLanguage(), grammar);
	}

	public double cosSim(Document document1, Document document2, boolean grammar) {
		if (document1.getLanguage() == null || !document1.getLanguage().equals(document2.getLanguage())) {
			throw new IllegalArgumentException("Cannot compare two documents of different languages.");
		}
		double sum = .0;
		for (Map.Entry<Term, Double> entry : document1.getNormalisedVector().entrySet()) {
			sum += tfIdf(entry.getKey(), document1, grammar) * tfIdf(entry.getKey(), document2, grammar);
		}
		//System.out.println(document1 + " " + document2 + " " + sum);
		return Math.sqrt(sum);
	}

	private void dfFromDb(Term term, Language language, boolean grammar) {
		String sql = "SELECT COUNT(DISTINCT(ct.corpus_id)) AS count FROM term AS t,corpus_term AS ct WHERE t.id=ct.term_id AND t.lemma=? AND t.language=?" + (grammar ? " AND t.pos=?" : "") + ";";
		try {
			PreparedStatement ps = database.getConnection().prepareStatement(sql);
			ps.setString(1, term.lemma);
			ps.setString(2, language.toString());
			if (grammar) {
				ps.setString(3, term.pos);
			}
			ResultSet rs = ps.executeQuery();
			if (rs.next()) {
				dfs.get(language).put(term, (double) rs.getInt(1));
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}
	}
}
